#include <Astri_DAQ.h>
#include <string>
#include <sstream>
#include <iomanip>
#include <unistd.h>
#include <boost/algorithm/string.hpp>

#include <Astri_Redis_DAQ_Observer.h>
#include <TCP_Protocol.h>
#include <UDP_Protocol.h>
#include <File_Archiver.h>
#include <Kafka_Provider.h>
#include <Redis_Provider.h>
#include <Astri_Monitor.h>
#include <Astri_MA_Generic.h>
#include <Packet_Structure_Json.h>

using namespace inaf::oasbo;
using namespace inaf::oasbo::DAQ;
using namespace inaf::oasbo::DAQ_observers;

DAQ::AstriDAQ::AstriDAQ() : tel_id(0), origin_id(0), prog_id(0), obs_id(0), run_id(0), data_type(0){
	this->protocol = ConnectionProtocols::TCPProtocolBuilder().getProtocol();
	this->archiver = Archivers::FileArchiverBuilder().getArchiver();
	this->provider = Providers::RedisProviderBuilder().getProvider();
	this->monitor = PacketMonitors::AstriPacketMonitorBuilder().getMonitor();

	std::string packets_path(std::getenv("BIAS_PREFIX"));
	packets_path.append("/config/packets/definitions/");
	std::string ma_packet(
			std::string(packets_path).append(
					"AstriMA/AstriMAGeneric.json"));

	Packets::PacketStructureJson structure(ma_packet);
	this->packet = new Packets::AstriMaGeneric(structure);
	this->currentState = Status::STOP;
}

std::string DAQ::AstriDAQ::getStateStr(Status state) {
	switch (state) {
	case Status::STOP:
		return std::string("STOP");
	case Status::IDLE:
		return std::string("IDLE");
	case Status::RUN:
		return std::string("RUN");
	case Status::READY:
		return std::string("READY");
	default:
		return std::string("NONE");
	}
}

void DAQ::AstriDAQ::stop() {
	this->archiver->close();
	this->provider->close();
	this->protocol->closeConnectionToClient();
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->stop();
			});
}

void DAQ::AstriDAQ::incrementRunID() {
	this->run_id += 1;
	std::map<std::string, std::string> newConf;
	for(auto conf : this->configurations){
		newConf[AstriDAQBuilder::run_id_key] = std::to_string(this->run_id);
		conf->insert(newConf, AstriDAQBuilder::config_target);
		conf->pushConfigToSource(AstriDAQBuilder::config_target);
	}
}

void DAQ::AstriDAQ::switchState(Status newState) {
	if (newState == STOP) {
		this->stop();
		this->setCurrentState(STOP);
		return;
	}

	std::string dest;
	switch (this->currentState) {
	case Status::READY: {
		switch (newState) {
		case Status::IDLE:
			this->setCurrentState(newState);
			break;
		default:
			std::cerr << "Switch from " << getStateStr(this->currentState)
					<< " to " << getStateStr(newState) << " not allowed"
					<< std::endl;
			break;
		}
		break;
	}
	case Status::IDLE: {
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			this->incrementRunID();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			break;
		case Status::RUN:
			archiver->close();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			break;
		case Status::READY:
			archiver->close();
			this->setCurrentState(Status::READY);
			protocol->connectToClient();
			if (!protocol->isConnectedToClient()) {
				std::cerr << "Unable to establish connection" << std::endl;
				switchState(Status::STOP);
			}
			incrementRunID();
			dest = computeNewDestination(true);
			archiver->setDest(dest);
			archiver->open();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::RUN: {
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			this->incrementRunID();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			break;
		case Status::RUN:
			archiver->close();
			this->incrementRunID();
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			break;
		case Status::READY:
			archiver->close();
			this->setCurrentState(Status::READY);
			protocol->connectToClient();
			if (!protocol->isConnectedToClient()) {
				std::cerr << "Unable to establish connection" << std::endl;
				switchState(Status::STOP);
			}
			incrementRunID();
			dest = computeNewDestination(true);
			archiver->setDest(dest);
			archiver->open();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::STOP:
		switch (newState) {
		case Status::READY:
			protocol->connectToClient();
			if (!protocol->isConnectedToClient()) {
				std::cerr << "Unable to establish connection" << std::endl;
				switchState(Status::STOP);
			}
			dest = computeNewDestination(true);
			archiver->setDest(dest);
			archiver->open();
			provider->open();
			this->setCurrentState(Status::READY);
			this->switchState(IDLE);
			break;
		default:
			std::cerr << "Switch from " << getStateStr(this->currentState)
					<< " to " << getStateStr(newState) << " not allowed"
					<< std::endl;
			break;
		}
		break;

	}
}

void DAQ::AstriDAQ::start() {
	incrementRunID();
	switchState(Status::READY);
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->start();
			}); // start observers
	// start receiving
	while (currentState != Status::STOP && currentState != Status::READY) {
		int rec = protocol->receiveFromClient(*this->packet);
		if (rec <= 0) { // error in receiving or closed connection
			if (!protocol->isConnectedToClient()) { // reconnect
				std::cout << "Connessione chiusa" << std::endl;
				this->switchState(Status::READY);
			}
			continue;
		}
		if (!this->packet->isRecognizedHeader()) {
			std::cerr << "Error: Packet header not recognized: "
					<< packet->operator []("type").value() << "-"
					<< packet->operator []("subtype").value() << "-"
					<< packet->operator []("length").value() << "-"
					<< std::endl;
			continue;
		}
		this->deliverPacket();

		if (this->packet->operator []("type") == 15) { //NOTIF PACKET
			monitor->printStats();
			unsigned short cmd = ((this->packet->getBinaryPointer()[8]) << 8)
					+ (this->packet->getBinaryPointer()[9]);
			Status newState = cmd == 43690 ? Status::RUN : Status::IDLE;
			this->setNextState(newState);
			std::cout << "Switching from " << getStateStr(this->currentState)
					<< " to -> " << getStateStr(newState) << std::endl;
			this->switchState(getNextState());
			this->updateObservers();
		}
	}
}
int DAQ::AstriDAQ::deliverPacket() {	//-------- CHECK CORRECTNESS
	archiver->write(*packet);
//	provider->write(*packet);
	monitor->monit(*packet);
	return 0;
}

void DAQ::AstriDAQ::updateObservers() {
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->updateAll();
			});
}

std::string DAQ::AstriDAQ::computeNewDestination(bool firstIdle) {
	std::stringstream ss;
	std::string mode_id;
	if (firstIdle)
		mode_id = std::string("F");
	else
		mode_id =
				this->currentState == Status::IDLE ?
						std::string("I") : std::string("R");
	ss << "astri_";
	ss << std::setfill('0') << std::setw(3) << tel_id << "_";
	ss << std::setfill('0') << std::setw(2) << origin_id << "_";
	ss << std::setfill('0') << std::setw(3) << prog_id << "_";
	ss << std::setfill('0') << std::setw(5) << obs_id << "_";
	ss << mode_id << "_";
	ss << std::setfill('0') << std::setw(6) << run_id << "_";
	ss << data_type;
	return ss.str();
}

