#include <Astri_DAQ.h>

#include <TCP_Protocol.h>
#include <UDP_Protocol.h>
#include <File_Archiver.h>
#include <Kafka_Provider.h>
#include <Redis_Provider.h>
#include <Astri_Monitor.h>
#include <Astri_MA_Generic.h>
#include <Packet_Structure_Json.h>

using namespace inaf::oasbo::DAQ;
using namespace inaf::oasbo;

AstriDAQBuilder::AstriDAQBuilder() {
	this->reset();
}
AstriDAQBuilder::~AstriDAQBuilder() {
	delete daq;
}
void AstriDAQBuilder::reset() {
	this->daq = new AstriDAQ();
}


AstriDAQBuilder* AstriDAQBuilder::configFrom(Configurators::BaseConfigurator &conf) {
	daq->configurations.push_back(&conf);
	std::string target = std::string(this->config_target);
	conf.readConfigFromSource(target);
	std::map<std::string, std::string> params = conf.getConfig();

// ---- PROTOCOL --------------------------------------------------------------------------
	std::string key = target+"_"+protocol_key;
	if (params.count(key) > 0) {
		if (params[key].compare(tcp_prot_key) == 0){
			daq->protocol =
					ConnectionProtocols::TCPProtocolBuilder().configFrom(conf)->getProtocol();
		}
		else if (params[key].compare(udp_prot_key) == 0) {
			daq->protocol =
					ConnectionProtocols::UDPProtocolBuilder().configFrom(conf)->getProtocol();
		}

		else { // not recognized, default to tcp
			daq->protocol = ConnectionProtocols::TCPProtocolBuilder().getProtocol();
		}
	}
// ---- ARCHIVER -----------------------------------------------------------------------------
	key = target+"_"+archiver_key;
	if (params.count(key) > 0) {
		if (params[key].compare(file_arch_key) == 0) {
			daq->archiver =
					Archivers::FileArchiverBuilder().configFrom(conf)->getArchiver();
		}

		else { // not recognized, default to file
			daq->archiver = Archivers::FileArchiverBuilder().getArchiver();
		}
	}
// ---- PROVIDER --------------------------------------------------------------------
	key = target+"_"+provider_key;
	if (params.count(key) > 0) {
		if (params[key].compare(kafka_prov_key) == 0) {
			daq->provider =
					Providers::KafkaProviderBuilder().configFrom(conf)->getProvider();
		} else if (params[key].compare(redis_prov_key) == 0) {
			daq->provider =
					Providers::RedisProviderBuilder().configFrom(conf)->getProvider();
		}
		else { // not recognized, default to  Redis
			daq->provider = Providers::RedisProviderBuilder().getProvider();
		}
	}
// ---- MONITOR -----------------------------------------------------------------------------
	key = target + "_" + monitor_key;
	if (params.count(key) > 0) {
		if (params[key].compare(astri_simple_monitor_key) == 0) {
			daq->monitor =
					PacketMonitors::AstriPacketMonitorBuilder().configFrom(conf)->getMonitor();
		}
		else { // not recognized, default to file
			daq->monitor =
					PacketMonitors::AstriPacketMonitorBuilder().configFrom(conf)->getMonitor();
		}
	}


	key = target+"_"+tel_id_key;
	if (params.count(key) > 0)
		daq->tel_id = std::stoi(params[key]);

	key = target+"_"+origin_id_key;
	if (params.count(key) > 0)
		daq->origin_id = std::stoi(params[key]);

	key = target+"_"+prog_id_key;
	if (params.count(key) > 0)
		daq->prog_id = std::stoi(params[key]);

	key = target+"_"+obs_id_key;
	if (params.count(key) > 0)
		daq->obs_id = std::stoi(params[key]);

	key = target+"_"+run_id_key;
	if (params.count(key) > 0)
		daq->run_id = std::stoi(params[key]);

	key = target+"_"+data_type_key;
	if (params.count(key) > 0)
		daq->data_type = std::stoi(params[key]);
	return this;
}

AstriDAQBuilder* AstriDAQBuilder::setProtocol(ConnectionProtocols::BaseProtocol *value) {
	this->daq->protocol = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setArchiver(Archivers::BaseArchiver *value) {
	this->daq->archiver = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setProvider(Providers::BaseProvider *value) {
	this->daq->provider = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setMonitor(PacketMonitors::BasePacketMonitor *value) {
	this->daq->monitor = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setPacket(PacketLib::BasePacket *value) {
	this->daq->packet = value;
	return this;
}

AstriDAQBuilder* AstriDAQBuilder::setTelId(int value) {
	this->daq->tel_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setOriginId(int value) {
	this->daq->origin_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setProgId(int value) {
	this->daq->prog_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setObsId(int value) {
	this->daq->obs_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setRunId(int value) {
	this->daq->run_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setDataType(int value) {
	this->daq->data_type = value;
	return this;
}

AstriDAQ* AstriDAQBuilder::getDAQ() {
	AstriDAQ *result = this->daq;
	this->reset();
	return result;
}

const std::string AstriDAQBuilder::config_target = "astridaq";

const std::string AstriDAQBuilder::protocol_key = "protocol";
const std::string AstriDAQBuilder::archiver_key = "archiver";
const std::string AstriDAQBuilder::provider_key = "provider";
const std::string AstriDAQBuilder::packet_key = "packet";
const std::string AstriDAQBuilder::monitor_key = "monitor";

const std::string AstriDAQBuilder::tcp_prot_key = "tcp";
const std::string AstriDAQBuilder::udp_prot_key = "udp";
const std::string AstriDAQBuilder::file_arch_key = "file";
const std::string AstriDAQBuilder::kafka_prov_key = "kafka";
const std::string AstriDAQBuilder::redis_prov_key = "redis";
const std::string AstriDAQBuilder::astri_ma_packet_key = "astri_ma";
const std::string AstriDAQBuilder::astri_horn_packet_key = "astri_horn";
const std::string AstriDAQBuilder::astri_simple_monitor_key = "astrisimple";

const std::string AstriDAQBuilder::tel_id_key = "tel_id";
const std::string AstriDAQBuilder::origin_id_key = "origin_id";
const std::string AstriDAQBuilder::prog_id_key = "prog_id";
const std::string AstriDAQBuilder::obs_id_key = "obs_id";
const std::string AstriDAQBuilder::run_id_key = "run_id";
const std::string AstriDAQBuilder::data_type_key = "data_type";

