/*
 * Astri_Redis_DAQ_Observer.cpp
 *
 *  Created on: Apr 5, 2023
 *      Author: valerio
 */

#include <Astri_Redis_DAQ_Observer.h>
#include <unistd.h>
#include <chrono>

using namespace inaf::oasbo::DAQ_observers;

AstriRedisDAQObserver::AstriRedisDAQObserver(inaf::oasbo::DAQ::BaseDAQ *daq, std::string ip, int port) :
		inaf::oasbo::DAQ_observers::BaseDAQ_Observer(*daq){
	this->updateInterval = 2;
	this->ip = ip;
	this->port = port;
	this->context = redisConnect(ip.c_str(), port);
	if (context == NULL || context->err) {
		if (context) {
			printf("Redis Observer Error: %s\n", context->errstr);
		} else {
			printf("Redis Observer Error: Can't allocate redis context\n");
		}
		context = nullptr;
	}
}

AstriRedisDAQObserver::~AstriRedisDAQObserver() {
	BaseDAQ_Observer::~BaseDAQ_Observer();
	redisFree(context);
}

void AstriRedisDAQObserver::updatePacketStats() {
	if(context == nullptr) return;
	inaf::oasbo::PacketMonitors::BasePacketMonitor &monitor =
			*(this->dataAcquisition->getMonitorPtr());
	for (const auto& [key, value] : monitor.getStatsMap()) {
		redisReply *r = (redisReply*) redisCommand(context, "SET %s %s",
				key.c_str(), value.c_str());
		if (r == NULL) {
			std::cerr << "Redis Observer Error executing updatePacketStats command"
					<< std::endl;
			return;
		}
		// Check the type of the Redis reply
		if (r->type == REDIS_REPLY_ERROR) {
			// Error executing Redis command
			std::cerr << r->str << std::endl;
			freeReplyObject(r);
			return;
		}
		freeReplyObject(r);
	}
}

void AstriRedisDAQObserver::updateArchiverStats() {
	if(context == nullptr) return;
	std::string dest = this->dataAcquisition->getArchiverPtr()->getDest();
	redisReply *r = (redisReply*) redisCommand(context,
			"SET LAST_CLOSED_FILE %s", dest.c_str());
	if (r == NULL) {
		std::cerr << "Redis Observer Error executing updateArchiverStats command" << std::endl;
		return;
	}
	// Check the type of the Redis reply
	if (r->type == REDIS_REPLY_ERROR) {
		// Error executing Redis command
		std::cerr << r->str << std::endl;
		freeReplyObject(r);
		return;
	}
	freeReplyObject(r);
}

void AstriRedisDAQObserver::updateProviderStats() {
	if(context == nullptr) return;
	std::string dest = this->dataAcquisition->getProviderPtr()->getDest();
	redisReply *r = (redisReply*) redisCommand(context,
			"SET PROVIDER_DESTINATION %s", dest.c_str());
	if (r == NULL) {
		std::cerr << "Redis Observer Error executing updateArchiverStats command" << std::endl;
		return;
	}
	// Check the type of the Redis reply
	if (r->type == REDIS_REPLY_ERROR) {
		// Error executing Redis command
		std::cerr << r->str << std::endl;
		freeReplyObject(r);
		return;
	}
	freeReplyObject(r);
}

void AstriRedisDAQObserver::updateReceiverStats() {
	if(context == nullptr) return;
	std::string dest = this->dataAcquisition->getSocketPtr()->getHost();
	redisReply *r = (redisReply*) redisCommand(context, "SET RECEIVER_HOST %s",
			dest.c_str());
	if (r == NULL) {
		std::cerr << "Redis Observer Error executing updateArchiverStats command" << std::endl;
		return;
	}
	// Check the type of the Redis reply
	if (r->type == REDIS_REPLY_ERROR) {
		// Error executing Redis command
		std::cerr << r->str << std::endl;
		freeReplyObject(r);
		return;
	}
	freeReplyObject(r);
}

void AstriRedisDAQObserver::updateAll() {
	if(context == nullptr) return;
	std::unique_lock < std::mutex
			> lock(this->writeStatsMutex, std::try_to_lock); // try to acquire the write mutex
	if (!lock.owns_lock()) {
		// mutex not acquired, exit function
		return;
	}
	updateArchiverStats();
	updatePacketStats();
	updateProviderStats();
	updateReceiverStats();
	lock.unlock(); // release
}

void AstriRedisDAQObserver::updateThread() {
	this->stopFlag = false;
	while (!this->stopFlag) {
		updateAll();
		std::unique_lock < std::mutex > sleep_lock(this->sleepMutex); // acquire sleep mutex
		this->sleep_cv.wait_for(sleep_lock, std::chrono::seconds(updateInterval)); // wait for "updateInterval" seconds or until notified
	}
}

void AstriRedisDAQObserver::start() {
	if (this->observerThread.joinable()) {
		return;
	}
	this->observerThread = std::thread(&AstriRedisDAQObserver::updateThread,
			this);
}

void AstriRedisDAQObserver::stop() {
	this->stopFlag = true;
	if (this->observerThread.joinable()) {
		this->sleep_cv.notify_all();
		this->observerThread.join();
	}
}
